#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === Extracter::Base
#
# This is the base class for the extracter-gem.
#
# Usage example:
#
#   Extracter::Base.new(ARGV)
#
# =========================================================================== #
# require 'extracter/base/base.rb'; < ::Extracter::Base
# =========================================================================== #
module Extracter

class Base # === Extracter::Base

  require 'fileutils' # We always need fileutils.

  begin
    require 'colours/html_colours'
  rescue LoadError; end

  require 'extracter/toplevel_methods/toplevel_methods.rb'

  # ========================================================================= #
  # === reset                                                     (reset tag)
  # ========================================================================= #
  def reset
    # ======================================================================= #
    # === @internal_hash
    #
    # The internal Hash must be created first.
    # ======================================================================= #
    create_the_internal_hash
    # ======================================================================= #
    # === :debug
    # ======================================================================= #
    @internal_hash[:debug] = false
    # ======================================================================= #
    # === :be_verbose
    # ======================================================================= #
    @internal_hash[:be_verbose] = true
    # ======================================================================= #
    # === :colour_to_use_for_directories
    # ======================================================================= #
    @internal_hash[:colour_to_use_for_directories] = 'cyan'
    # ======================================================================= #
    # === :use_opn
    # ======================================================================= #
    @internal_hash[:use_opn] = true # ← Whether to use make use of Opn by default or not.
    # ======================================================================= #
    # Next check whether the Colours gem is available.
    # ======================================================================= #
    check_whether_the_colours_gem_is_available
    do_try_to_use_colours
  end

  # ========================================================================= #
  # === create_the_internal_hash
  # ========================================================================= #
  def create_the_internal_hash
    @internal_hash = {}
  end

  # ========================================================================= #
  # === check_whether_the_colours_gem_is_available
  # ========================================================================= #
  def check_whether_the_colours_gem_is_available
    if Object.const_defined?(:Colours) and
       Colours.respond_to?(:simp) and
       Colours.respond_to?(:sfancy) and
       Colours.respond_to?(:sdir) and
       Colours.respond_to?(:simp)
    else
      do_not_make_use_of_colours
    end
  end

  # ========================================================================= #
  # === do_not_make_use_of_colours
  # ========================================================================= #
  def do_not_make_use_of_colours
    @internal_hash[:use_colours] = false
  end

  # ========================================================================= #
  # === change_directory
  # ========================================================================= #
  def change_directory(i)
    Dir.chdir(i)
  end; alias cd change_directory # === cd

  # ========================================================================= #
  # === cyan
  # ========================================================================= #
  def cyan?(use_colours = use_colours?)
    return Colours::CYAN if Object.const_defined?(:Colours) and use_colours
    return ''
  end

  # ========================================================================= #
  # === tomato
  # ========================================================================= #
  def tomato(
      i           = '',
      use_colours = use_colours?
    )
    return Colours.tomato(i) if use_colours
    return i
  end

  # ========================================================================= #
  # === simp
  # ========================================================================= #
  def simp(
      i           = '',
      use_colours = use_colours?
    )
    return Colours.simp(i) if use_colours
    return i
  end

  # ========================================================================= #
  # === steelblue
  # ========================================================================= #
  def steelblue(
      i           = '',
      use_colours = use_colours?
    )
    return ::Colours.steelblue(i) if use_colours and ::Colours.respond_to?(:steelblue)
    return i
  end

  # ========================================================================= #
  # === ewarn
  # ========================================================================= #
  def ewarn(
      i           = '',
      use_colours = use_colours?
    )
    if use_colours and Object.const_defined?(:Colours)
      e Colours.swarn(i)
    else
      e i
    end
  end

  # ========================================================================= #
  # === cpr
  # ========================================================================= #
  def cpr(a, b)
    try_to_require_fileutils
    FileUtils.cp_r(a, b)
  end

  # ========================================================================= #
  # === move
  # ========================================================================= #
  def move(a, b)
    try_to_require_fileutils
    FileUtils.mv(a, b)
  end; alias mv move # === mv

  # ========================================================================= #
  # === remove_this_directory
  #
  # This method can be used to remove a directory.
  # ========================================================================= #
  def remove_this_directory(i)
    if File.directory? i
      try_to_require_fileutils
      FileUtils.rm_r(i) unless i == '/'
    end
  end

  # ========================================================================= #
  # === try_to_require_fileutils
  # ========================================================================= #
  def try_to_require_fileutils
    require 'fileutils' unless Object.const_defined?(:FileUtils)
  end

  # ========================================================================= #
  # === sfile
  # ========================================================================= #
  def sfile(i = '')
    return Colours.sfancy(i) if use_colours?
    return i
  end

  # ========================================================================= #
  # === sfancy
  # ========================================================================= #
  def sfancy(i = '')
    return Colours.sfancy(i) if use_colours?
    return i
  end

  # ========================================================================= #
  # === sdir
  # ========================================================================= #
  def sdir(i = '')
    return Colours.sdir(i) if use_colours?
    return i
  end

  # ========================================================================= #
  # === e                                                             (e tag)
  # ========================================================================= #
  def e(i = '')
    puts i
  end

  # ========================================================================= #
  # === rev                                                         (rev tag)
  # ========================================================================= #
  def rev
    if use_colours?
      ::Colours.rev
    else
      ''
    end
  end

  # ========================================================================= #
  # === are_we_on_windows?
  # ========================================================================= #
  def are_we_on_windows?
    ::Extracter.are_we_on_windows?
  end
 
  # ========================================================================= #
  # === be_verbose?
  #
  # Getter method for whether we will be verbose or not.
  # ========================================================================= #
  def be_verbose?
    @internal_hash[:be_verbose]
  end

  # ========================================================================= #
  # === register_sigint
  # ========================================================================= #
  def register_sigint(this_class_name = 'Extracter')
    Signal.trap('SIGINT') {
      e sfancy('Requesting a graceful exit from ')+
        colour_to_use_for_directories?+
        'class '+this_class_name+sfancy('. Exiting now.')
      exit
    }
  end

  # ========================================================================= #
  # === colour_to_use_for_directories?
  # ========================================================================= #
  def colour_to_use_for_directories?
    if use_colours?
      return @internal_hash[:colour_to_use_for_directories]
    else
      return ''
    end
  end

  # ========================================================================= #
  # === disable_colours
  #
  # Use this method if you want to disable colour-support of this class.
  # ========================================================================= #
  def disable_colours
    set_use_colours(false)
    @internal_hash[:colour_to_use_for_directories] = ''.dup
  end

  # ========================================================================= #
  # === create_directory                                          (mkdir tag)
  #
  # Use this to create directories.
  # ========================================================================= #
  def create_directory(i)
    FileUtils.mkdir_p(i) unless File.directory?(i)
  end; alias mkdir create_directory # === mkdir

  # ========================================================================= #
  # === be_verbose
  # ========================================================================= #
  def be_verbose
    set_be_verbose(true)
  end

  # ========================================================================= #
  # === do_be_quiet
  # ========================================================================= #
  def do_be_quiet
    set_be_verbose(false)
  end; alias be_silent do_be_quiet # === be_silent
       alias be_quiet  do_be_quiet # === be_quiet

  # ========================================================================= #
  # === rds
  # ========================================================================= #
  def rds(i)
    return i.squeeze('/') if i.respond_to? :squeeze
    return i
  end

  # ========================================================================= #
  # === set_be_verbose
  #
  # This sets the verbosity level of the class. Use only this method
  # when you wish to modify the @be_verbose instance variable.
  # ========================================================================= #
  def set_be_verbose(i = false)
    @internal_hash[:be_verbose] = i
  end; alias set_verbosity      set_be_verbose # === set_verbosity
       alias show_commands_used set_be_verbose # === show_commands_used
  
  # ========================================================================= #
  # === remove_file_extension
  # ========================================================================= #
  def remove_file_extension(i)
    _ = File.basename(i)
    return ::Extracter.remove_archive_type(_)
  end; alias remove_extension remove_file_extension # === remove_extensions
       alias remove_ext       remove_file_extension # === remove_ext

  # ========================================================================= #
  # === set_commandline_arguments
  # ========================================================================= #
  def set_commandline_arguments(i = '')
    i = [i].flatten.compact
    @commandline_arguments = i
  end

  # ========================================================================= #
  # === commandline_arguments?
  # ========================================================================= #
  def commandline_arguments?
    @commandline_arguments
  end

  # ========================================================================= #
  # === first_argument?
  # ========================================================================= #
  def first_argument?
    @commandline_arguments.first
  end; alias first? first_argument? # === first?

  # ========================================================================= #
  # === return_pwd
  # ========================================================================= #
  def return_pwd
    "#{Dir.pwd}/".squeeze('/')
  end

  # ========================================================================= #
  # === debug?
  # ========================================================================= #
  def debug?
    @internal_hash[:debug]
  end

  # ========================================================================= #
  # === enable_debug
  # ========================================================================= #
  def enable_debug
    @internal_hash[:debug] = true
  end

  # ========================================================================= #
  # === namespace?
  # ========================================================================= #
  def namespace?
    @internal_hash[:namespace]
  end

  # ========================================================================= #
  # === enable_colours
  # ========================================================================= #
  def enable_colours
    set_use_colours(true)
    @internal_hash[:colour_to_use_for_directories] = cyan?
  end

  # ========================================================================= #
  # === set_use_colours
  # ========================================================================= #
  def set_use_colours(i, main_hash = main_hash?)
    # ======================================================================= #
    # We must also sync this towards our main Hash, for opn(). The next
    # line of code achieves precisely that.
    # ======================================================================= #
    main_hash.update(use_colours: i) if main_hash
    @internal_hash[:try_to_use_colours] = i
  end

  # ========================================================================= #
  # === do_try_to_use_colours
  # ========================================================================= #
  def do_try_to_use_colours
    set_use_colours(true)
  end

  # ========================================================================= #
  # === try_to_use_colours?
  #
  # Determine whether we will use colours for class Extracter::Base
  # and its subclass.
  # ========================================================================= #
  def try_to_use_colours?
    @internal_hash[:try_to_use_colours]
  end; alias use_colours? try_to_use_colours? # === use_colours?
       alias colours?     try_to_use_colours? # === colours?

  # ========================================================================= #
  # === set_use_opn
  # ========================================================================= #
  def set_use_opn(i = true)
    @internal_hash[:use_opn] = i
  end

  # ========================================================================= #
  # === use_opn?
  # ========================================================================= #
  def use_opn?
    @internal_hash[:use_opn]
  end

  # ========================================================================= #
  # === opne
  # ========================================================================= #
  def opne(i = '')
    opnn; e i
  end

  # ========================================================================= #
  # === use_this_opn_hash?
  # ========================================================================= #
  def use_this_opn_hash?
    @internal_hash[:use_this_opn_hash]
  end; alias main_hash? use_this_opn_hash? # === main_hash?

  # ========================================================================= #
  # === opnn
  #
  # This variant will also check whether we should show the name or not.
  # ========================================================================= #
  def opnn(
      use_this_hash = use_this_opn_hash?
    )
    if use_opn? and Object.const_defined?(:Opn)
      Opn.opn(use_this_hash)
    end
  end; alias opn  opnn # === opn
       alias copn opnn # === copn

  # ========================================================================= #
  # === is_archive?
  #
  # This method queries whether the given file at hand (as input to
  # this method) is assumed to be an archive or whether it is not.
  # ========================================================================= #
  def is_archive?(i)
    ARRAY_ARCHIVE_TYPES.include?(
      File.extname(File.basename(i))
    )
  end

  # ========================================================================= #
  # === array_archive_types?
  # ========================================================================= #
  def array_archive_types?
    ARRAY_REGISTERED_ARCHIVES
  end

  # ========================================================================= #
  # === orev
  # ========================================================================= #
  def orev(i = '')
    opne "#{rev}#{i}"
  end

end; end

if __FILE__ == $PROGRAM_NAME
  Extracter::Extracter.new(ARGV)
end # base.rb
